/*
 * prop_dialog_action.c  -- action dialog to edit action
 *
 * $Id: prop_dialog_action.c,v 1.8 2004/03/05 07:43:21 hos Exp $
 *
 */

#include <windows.h>
#include "main.h"
#include "util.h"


/* prototypes */
static void action_dlg_update_page(prop_page_context_t *ctx);
static void action_dlg_joy_update_val(prop_page_context_t *ctx);


/* page proc alist */
struct uint_ptr_pair action_dlg[];
static struct uint_ptr_pair command_proc_map[];

static struct uint_ptr_pair action_dlg_mod[];

static struct uint_ptr_pair action_dlg_delay[];

static struct uint_ptr_pair action_dlg_joy[];
static struct uint_ptr_pair joy_command_proc_map[];

static struct uint_ptr_pair action_dlg_apply[];


/* implementations */

/* WM_INITDIALOG for main */
static INT_PTR action_dlg_init_dialog(HWND hwnd, UINT msg,
                                      WPARAM wparam, LPARAM lparam,
                                      prop_page_context_t *ctx)
{
    int i;
    struct uint_ptr_pair id_item_map[] = {
        {IDC_CMB_ACTION_TYPE, &ctx->action_dlg.main.type},

        {0, NULL}
    };

    get_dialog_item(hwnd, id_item_map);
    ctx->command_proc = command_proc_map;

    ctx->action_dlg.self = hwnd;

    {
        HWND place_holder;
        RECT rect;
        POINT pt;

        struct {
            int id;
            struct uint_ptr_pair *page_proc;
            UCHAR type;
            LPCTSTR str_type;
        } page[] = {
            {IDD_ACTION_EDIT_JOY, action_dlg_joy,
             CONTROLLER_ACTION_JOY, TEXT("Joystick")},
            {IDD_ACTION_EDIT_MOD, action_dlg_mod,
             CONTROLLER_ACTION_MOD, TEXT("Modifier")},
            {IDD_ACTION_EDIT_DELAY, action_dlg_delay,
             CONTROLLER_ACTION_DELAY, TEXT("Delay")},
            {IDD_ACTION_EDIT_APPLY, action_dlg_apply,
             CONTROLLER_ACTION_APPLY, TEXT("Action list")},

            {0, NULL}
        };

        /* get window place from place holder */
        place_holder = GetDlgItem(hwnd, IDC_ACTION_EDIT);
        GetWindowRect(place_holder, &rect);

        pt.x = rect.left;
        pt.y = rect.top;
        ScreenToClient(hwnd, &pt);

        /* count num of pages */
        for(i = 0; page[i].page_proc != NULL; i++) ;
        ctx->action_dlg.main.num_page = i;

        /* allocate page data */
        ctx->action_dlg.main.page_ctx = 
            (prop_page_context_t *)malloc(sizeof(prop_page_context_t) *
                                          ctx->action_dlg.main.num_page);
        if(ctx->action_dlg.main.page_ctx == NULL) {
            return FALSE;
        }
        memset(ctx->action_dlg.main.page_ctx, 0,
               sizeof(prop_page_context_t) * ctx->action_dlg.main.num_page);

        for(i = 0; page[i].page_proc != NULL; i++) {
            /* create page window */
            memcpy(&ctx->action_dlg.main.page_ctx[i], ctx,
                   sizeof(prop_page_context_t));
            memset(&ctx->action_dlg.main.page_ctx[i].action_dlg, 0,
                   sizeof(ctx->action_dlg.main.page_ctx[i].action_dlg));
            memcpy(&ctx->action_dlg.main.page_ctx[i].action_dlg.action_conf,
                   &ctx->action_dlg.action_conf,
                   sizeof(struct controller_action_config));
            ctx->action_dlg.main.page_ctx[i].page_proc = page[i].page_proc;
            ctx->action_dlg.main.page_ctx[i].action_dlg.self =
                CreateDialogParam(instance, MAKEINTRESOURCE(page[i].id),
                                  hwnd, prop_subpage_proc,
                                  (LPARAM)&ctx->action_dlg.main.page_ctx[i]);

            /* hide */
            ShowWindow(ctx->action_dlg.main.page_ctx[i].action_dlg.self,
                       SW_HIDE);

            /* move */
            MoveWindow(ctx->action_dlg.main.page_ctx[i].action_dlg.self,
                       pt.x, pt.y,
                       rect.right - rect.left, rect.bottom - rect.top,
                       FALSE);

            /* add to combobox */
            SendMessage(ctx->action_dlg.main.type, CB_INSERTSTRING,
                        i, (LPARAM)page[i].str_type);

            if(ctx->action_dlg.action_conf.type == page[i].type) {
                /* select if type match */
                SendMessage(ctx->action_dlg.main.type, CB_SETCURSEL, i, 0);
            }
        }

        /* destroy place holder */
        DestroyWindow(place_holder);
    }

    if(SendMessage(ctx->action_dlg.main.type, CB_GETCURSEL, 0, 0) == CB_ERR) {
        /* select first item if not yet selected */
        SendMessage(ctx->action_dlg.main.type, CB_SETCURSEL, 0, 0);
    }
    action_dlg_update_page(ctx);

    return TRUE;
}

/* WM_DESTROY */
static INT_PTR action_dlg_destroy(HWND hwnd, UINT msg,
                                  WPARAM wparam, LPARAM lparam,
                                  prop_page_context_t *ctx)
{
    int i;

    if(ctx->action_dlg.main.page_ctx != NULL) {
        for(i = 0; i < ctx->action_dlg.main.num_page; i++) {
            DestroyWindow(ctx->action_dlg.main.page_ctx[i].action_dlg.self);
        }

        free(ctx->action_dlg.main.page_ctx);
    }

    ctx->action_dlg.main.page_ctx = NULL;

    return TRUE;
}

/* BN_CLICK for ok */
static INT_PTR action_dlg_ok_clicked(HWND hwnd,
                                     WORD code, WORD id, HWND ctrl,
                                     prop_page_context_t *ctx)
{
    int idx;

    idx = SendMessage(ctx->action_dlg.main.type, CB_GETCURSEL, 0, 0);
    if(idx == CB_ERR) {
        return FALSE;
    }

    if(SendMessage(ctx->action_dlg.main.page_ctx[idx].action_dlg.self,
                   WM_ACTION_DLG_QUERY_VALID, 0, 0) ==
       FALSE) {
        return TRUE;
    }

    memcpy(&ctx->action_dlg.action_conf,
           &ctx->action_dlg.main.page_ctx[idx].action_dlg.action_conf,
           sizeof(struct controller_action_config));

    EndDialog(hwnd, TRUE);

    return TRUE;
}

/* BN_CLICK for cancel */
static INT_PTR action_dlg_cancel_clicked(HWND hwnd,
                                         WORD code, WORD id, HWND ctrl,
                                         prop_page_context_t *ctx)
{
    EndDialog(hwnd, FALSE);

    return TRUE;
}

/* CBN_SELCHANGE for type */
static INT_PTR action_dlg_type_change(HWND hwnd,
                                      WORD code, WORD id, HWND ctrl,
                                      prop_page_context_t *ctx)
{
    action_dlg_update_page(ctx);

    return TRUE;
}

static void action_dlg_update_page(prop_page_context_t *ctx)
{
    int i, idx;

    for(i = 0; i < ctx->action_dlg.main.num_page; i++) {
        ShowWindow(ctx->action_dlg.main.page_ctx[i].action_dlg.self, SW_HIDE);
    }

    idx = SendMessage(ctx->action_dlg.main.type, CB_GETCURSEL, 0, 0);
    if(idx == CB_ERR) {
        return;
    }

    ShowWindow(ctx->action_dlg.main.page_ctx[idx].action_dlg.self, SW_SHOW);
}


/* WM_INITDIALOG for mod */
static INT_PTR action_dlg_mod_init_dialog(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam,
                                          prop_page_context_t *ctx)
{
    struct uint_ptr_pair id_item_map[] = {
        {IDC_CHK_ACTION_MOD(1), &ctx->action_dlg.mod.mod[0]},
        {IDC_CHK_ACTION_MOD(2), &ctx->action_dlg.mod.mod[1]},
        {IDC_CHK_ACTION_MOD(3), &ctx->action_dlg.mod.mod[2]},
        {IDC_CHK_ACTION_MOD(4), &ctx->action_dlg.mod.mod[3]},
        {IDC_CHK_ACTION_MOD(5), &ctx->action_dlg.mod.mod[4]},
        {IDC_CHK_ACTION_MOD(6), &ctx->action_dlg.mod.mod[5]},
        {IDC_CHK_ACTION_MOD(7), &ctx->action_dlg.mod.mod[6]},
        {IDC_CHK_ACTION_MOD(8), &ctx->action_dlg.mod.mod[7]},

        {IDC_CHK_ACTION_MOD(9), &ctx->action_dlg.mod.mod[8]},
        {IDC_CHK_ACTION_MOD(10), &ctx->action_dlg.mod.mod[9]},
        {IDC_CHK_ACTION_MOD(11), &ctx->action_dlg.mod.mod[10]},
        {IDC_CHK_ACTION_MOD(12), &ctx->action_dlg.mod.mod[11]},
        {IDC_CHK_ACTION_MOD(13), &ctx->action_dlg.mod.mod[12]},
        {IDC_CHK_ACTION_MOD(14), &ctx->action_dlg.mod.mod[13]},
        {IDC_CHK_ACTION_MOD(15), &ctx->action_dlg.mod.mod[14]},
        {IDC_CHK_ACTION_MOD(16), &ctx->action_dlg.mod.mod[15]},

        {0, NULL}
    };

    get_dialog_item(hwnd, id_item_map);

    /* set checkbox */
    {
        USHORT mod, mod_mask;

        if(ctx->action_dlg.action_conf.type == CONTROLLER_ACTION_MOD) {
            mod = ctx->action_dlg.action_conf.mod.mod;
            mod_mask = ctx->action_dlg.action_conf.mod.mod_mask;
        } else {
            mod = 0x0000;
            mod_mask = 0x0000;
        }

        set_modifier_bits_to_checkbox(ctx->action_dlg.mod.mod, mod, mod_mask);
    }

    return TRUE;
}

/* WM_ACTION_DLG_QUERY_VALID for mod */
static INT_PTR action_dlg_mod_query_valid(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam,
                                          prop_page_context_t *ctx)
{
    ctx->action_dlg.action_conf.type = CONTROLLER_ACTION_MOD;

    get_modifier_bits_from_checkbox(ctx->action_dlg.mod.mod,
                                    &ctx->action_dlg.action_conf.mod.mod,
                                    &ctx->action_dlg.action_conf.mod.mod_mask);

    SetWindowLong(hwnd, DWL_MSGRESULT, TRUE);
    return TRUE;
}


/* WM_INITDIALOG for delay */
static INT_PTR action_dlg_delay_init_dialog(HWND hwnd, UINT msg,
                                            WPARAM wparam, LPARAM lparam,
                                            prop_page_context_t *ctx)
{
    struct uint_ptr_pair id_item_map[] = {
        {IDC_EDIT_ACTION_DELAY, &ctx->action_dlg.delay.msec},
        {IDC_SPIN_ACTION_DELAY, &ctx->action_dlg.delay.spin},

        {0, NULL}
    };

    get_dialog_item(hwnd, id_item_map);

    /* set range of spin control: 0s - 100s */
    SendMessage(ctx->action_dlg.delay.spin, UDM_SETRANGE32, 0, 1000 * 100);

    /* set pos of spin control */
    {
        LPARAM pos;

        if(ctx->action_dlg.action_conf.type == CONTROLLER_ACTION_DELAY) {
            pos = ctx->action_dlg.action_conf.delay.msec;
        } else {
            pos = 500;
        }

        SendMessage(ctx->action_dlg.delay.spin, UDM_SETPOS32, 0, pos);
    }

    return TRUE;
}

/* WM_ACTION_DLG_QUERY_VALID for delay */
static INT_PTR action_dlg_delay_query_valid(HWND hwnd, UINT msg,
                                            WPARAM wparam, LPARAM lparam,
                                            prop_page_context_t *ctx)
{
    LRESULT val;
    BOOL errp;

    ctx->action_dlg.action_conf.type = CONTROLLER_ACTION_DELAY;

    val = SendMessage(ctx->action_dlg.delay.spin, UDM_GETPOS32,
                      0, (LPARAM)&errp);
    if(errp) {
        display_message_box(IDS_MSG_ENTER_ACTION_DELAY_MSEC, IDS_MSG_ERROR,
                            hwnd, MB_OK | MB_ICONSTOP);
        SetFocus(ctx->action_dlg.delay.msec);

        SetWindowLong(hwnd, DWL_MSGRESULT, FALSE);
        return TRUE;
    }

    ctx->action_dlg.action_conf.delay.msec = val;

    SetWindowLong(hwnd, DWL_MSGRESULT, TRUE);
    return TRUE;
}


/* WM_INITDIALOG for joy */
static INT_PTR action_dlg_joy_init_dialog(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam,
                                          prop_page_context_t *ctx)
{
    int i;
    struct uint_ptr_pair id_item_map[] = {
        {IDC_CMB_ACTION_JOY_CODE, &ctx->action_dlg.joy.target},
        {IDC_CMB_ACTION_JOY_TYPE, &ctx->action_dlg.joy.type},

        {IDC_ACTION_JOY_VAL_SEP, &ctx->action_dlg.joy.sep},

        {IDC_EDIT_ACTION_JOY_VAL_RATIO, &ctx->action_dlg.joy.ratio},
        {IDC_SPIN_ACTION_JOY_VAL_RATIO, &ctx->action_dlg.joy.ratio_spin},

        {IDC_EDIT_ACTION_JOY_VAL_NUMER, &ctx->action_dlg.joy.numer},
        {IDC_SPIN_ACTION_JOY_VAL_NUMER, &ctx->action_dlg.joy.numer_spin},

        {IDC_EDIT_ACTION_JOY_VAL_DENOM, &ctx->action_dlg.joy.denom},
        {IDC_SPIN_ACTION_JOY_VAL_DENOM, &ctx->action_dlg.joy.denom_spin},

        {IDC_EDIT_ACTION_JOY_VAL_OFFST, &ctx->action_dlg.joy.offst},
        {IDC_SPIN_ACTION_JOY_VAL_OFFST, &ctx->action_dlg.joy.offst_spin},

        {0, NULL}
    };

    get_dialog_item(hwnd, id_item_map);
    ctx->command_proc = joy_command_proc_map;

    /* target combobox */
    {
        WCHAR buf[256];
        LPWSTR code;

        for(i = 0; i <= CONTROLLER_ACTION_JOY_CODE_MAX; i++) {
            code = (LPWSTR)assq_pair(action_code_map, i, NULL);
            if(code == NULL) {
                wsprintfW(buf, action_code_button_base,
                          i - CONTROLLER_ACTION_JOY_CODE_BTN(0) + 1);
                code = buf;
            }

            SendMessage(ctx->action_dlg.joy.target, CB_INSERTSTRING,
                        i, (LPARAM)code);
        }

        if(ctx->action_dlg.action_conf.type == CONTROLLER_ACTION_JOY) {
            SendMessage(ctx->action_dlg.joy.target, CB_SETCURSEL,
                        ctx->action_dlg.action_conf.joy.code, 0);
        }
    }

    /* type combobox */
    {
        struct uint_ptr_pair flag_str[] = {
            {CONTROLLER_ACTION_JOY_VAL_MIN, L"Minimum"},
            {CONTROLLER_ACTION_JOY_VAL_MAX, L"Maximum"},
            {CONTROLLER_ACTION_JOY_VAL_MDL, L"Center"},
            {CONTROLLER_ACTION_JOY_VAL_THROUGH, L"Pass through"},
            {CONTROLLER_ACTION_JOY_VAL_REVERSE, L"Reverse"},
            {CONTROLLER_ACTION_JOY_VAL_SPEC, L"Specified value"},
            {CONTROLLER_ACTION_JOY_VAL_RATIO, L"Calcutated value"},

            {0, NULL}
        };

        for(i = 0; flag_str[i].ptr != NULL; i++) {
            SendMessage(ctx->action_dlg.joy.type, CB_INSERTSTRING,
                        i, (LPARAM)flag_str[i].ptr);
            SendMessage(ctx->action_dlg.joy.type, CB_SETITEMDATA,
                        i, flag_str[i].key);

            if(ctx->action_dlg.action_conf.type == CONTROLLER_ACTION_JOY &&
               ctx->action_dlg.action_conf.joy.flags == flag_str[i].key) {
                SendMessage(ctx->action_dlg.joy.type, CB_SETCURSEL, i, 0);
            }
        }
    }

    {
        struct {
            HWND hwnd;
            int min;
            int max;
        } spin[] = {
            {ctx->action_dlg.joy.ratio_spin, 0, 255},
            {ctx->action_dlg.joy.numer_spin, -500, 500},
            {ctx->action_dlg.joy.denom_spin, 1, 500},
            {ctx->action_dlg.joy.offst_spin, -500, 500},

            {NULL, 0, 0}
        };

        for(i = 0; spin[i].hwnd != NULL; i++) {
            /* set range of spin control: 0 - 500 */
            SendMessage(spin[i].hwnd, UDM_SETRANGE32,
                        spin[i].min, spin[i].max);
        }
    }

    action_dlg_joy_update_val(ctx);

    return TRUE;
}

static HWND action_dlg_joy_get_value(prop_page_context_t *ctx)
{
    int i;
    BOOL errp;
    HWND err_hwnd;

    struct {
        HWND spin;
        USHORT flags;
        LRESULT val;
    } spin_map[] = {
        {ctx->action_dlg.joy.ratio_spin, CONTROLLER_ACTION_JOY_VAL_SPEC},
        {ctx->action_dlg.joy.numer_spin, CONTROLLER_ACTION_JOY_VAL_RATIO},
        {ctx->action_dlg.joy.denom_spin, CONTROLLER_ACTION_JOY_VAL_RATIO},
        {ctx->action_dlg.joy.offst_spin, CONTROLLER_ACTION_JOY_VAL_RATIO},

        {NULL, 0}
    };

    err_hwnd = NULL;
    for(i = 0; spin_map[i].spin != NULL; i++) {
        if(spin_map[i].flags == ctx->action_dlg.action_conf.joy.flags) {
            spin_map[i].val = SendMessage(spin_map[i].spin, UDM_GETPOS32,
                                          0, (LPARAM)&errp);
            if(errp) {
                int min, max;
                SendMessage(spin_map[i].spin, UDM_GETRANGE32,
                            (WPARAM)&min, (LPARAM)&max);
                spin_map[i].val = min;

                err_hwnd = spin_map[i].spin;
            }
        }
    }

    switch(ctx->action_dlg.action_conf.joy.flags) {
      case CONTROLLER_ACTION_JOY_VAL_MIN:
      case CONTROLLER_ACTION_JOY_VAL_MAX:
      case CONTROLLER_ACTION_JOY_VAL_MDL:
      case CONTROLLER_ACTION_JOY_VAL_THROUGH:
      case CONTROLLER_ACTION_JOY_VAL_REVERSE:
          break;

      case CONTROLLER_ACTION_JOY_VAL_SPEC:
          ctx->action_dlg.action_conf.joy.val = spin_map[0].val;
          break;

      case CONTROLLER_ACTION_JOY_VAL_RATIO:
          ctx->action_dlg.action_conf.joy.ratio_numerator =
              spin_map[1].val;
          ctx->action_dlg.action_conf.joy.ratio_denominator =
              spin_map[2].val;
          ctx->action_dlg.action_conf.joy.offset = spin_map[3].val;
          break;
    }

    return err_hwnd;
}

/* WM_ACTION_DLG_QUERY_VALID for joy */
static INT_PTR action_dlg_joy_query_valid(HWND hwnd, UINT msg,
                                          WPARAM wparam, LPARAM lparam,
                                          prop_page_context_t *ctx)
{
    int idx;

    ctx->action_dlg.action_conf.type = CONTROLLER_ACTION_JOY;

    idx = SendMessage(ctx->action_dlg.joy.target, CB_GETCURSEL, 0, 0);
    if(idx == CB_ERR) {
        display_message_box(IDS_MSG_SELECT_JOY_TARGET, IDS_MSG_ERROR,
                            hwnd, MB_OK | MB_ICONSTOP);
        SetFocus(ctx->action_dlg.joy.target);

        SetWindowLong(hwnd, DWL_MSGRESULT, FALSE);
        return TRUE;
    }
    ctx->action_dlg.action_conf.joy.code = idx;

    idx = SendMessage(ctx->action_dlg.joy.type, CB_GETCURSEL, 0, 0);
    if(idx == CB_ERR) {
        display_message_box(IDS_MSG_SELECT_JOY_TYPE, IDS_MSG_ERROR,
                            hwnd, MB_OK | MB_ICONSTOP);
        SetFocus(ctx->action_dlg.joy.type);

        SetWindowLong(hwnd, DWL_MSGRESULT, FALSE);
        return TRUE;
    }

    {
        HWND err_hwnd;

        err_hwnd = action_dlg_joy_get_value(ctx);
        if(err_hwnd != NULL) {
            display_message_box(IDS_MSG_ENTER_ACTIO_JOY_VALUE,
                                IDS_MSG_ERROR,
                                hwnd, MB_OK | MB_ICONSTOP);
            SetFocus(err_hwnd);

            SetWindowLong(hwnd, DWL_MSGRESULT, FALSE);
            return TRUE;            
        }
    }

    SetWindowLong(hwnd, DWL_MSGRESULT, TRUE);
    return TRUE;
}

/* WM_DRAWITEM */
static INT_PTR action_dlg_joy_drawitem(HWND hwnd, UINT msg,
                                       WPARAM wparam, LPARAM lparam,
                                       prop_page_context_t *ctx)
{
    DRAWITEMSTRUCT *item;

    item = (DRAWITEMSTRUCT *)lparam;

    if(item->itemAction != ODA_DRAWENTIRE) {
        return FALSE;
    }

    if(item->CtlID != IDC_ACTION_JOY_VAL_SEP) {
        return FALSE;
    }

    {
        HDC hdc;
        HPEN pen;
        HGDIOBJ old_pen;
        RECT rect;
        int xs, xe, y;

        hdc = item->hDC;

        GetClientRect(ctx->action_dlg.joy.sep, &rect);
        pen = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_BTNTEXT));

        old_pen = SelectObject(hdc, pen);

        xs = rect.left;
        xe = rect.right;
        y = (rect.top + rect.bottom) / 2;

        MoveToEx(hdc, xs, y, NULL);
        LineTo(hdc, xe, y);

        SelectObject(hdc, old_pen);

        DeleteObject(pen);
    }

    return TRUE;
}

/* TVN_SELCHANGED */
static INT_PTR action_dlg_joy_type_changed(HWND hwnd,
                                           WORD code, WORD id, HWND ctrl,
                                           prop_page_context_t *ctx)
{
    action_dlg_joy_get_value(ctx);
    action_dlg_joy_update_val(ctx);

    return TRUE;
}

static void set_window_num_value(HWND edit, HWND spin, int val)
{
    TCHAR buf[32];

    SendMessage(spin, UDM_SETPOS32, 0, val);

    wsprintf(buf, TEXT("%d"), val);
    SetWindowText(edit, buf);
}

static void action_dlg_joy_update_val(prop_page_context_t *ctx)
{
    int idx, val;
    int i;
    HWND edit[] = {
        ctx->action_dlg.joy.ratio,
        ctx->action_dlg.joy.numer,
        ctx->action_dlg.joy.denom,
        ctx->action_dlg.joy.offst,

        NULL
    };

    for(i = 0; edit[i] != NULL; i++) {
        EnableWindow(edit[i], FALSE);
        SetWindowText(edit[i], TEXT(""));
    }

    idx = SendMessage(ctx->action_dlg.joy.type, CB_GETCURSEL, 0, 0);
    if(idx == CB_ERR) {
        return;
    }

    val = SendMessage(ctx->action_dlg.joy.type, CB_GETITEMDATA, idx, 0);
    ctx->action_dlg.action_conf.joy.flags = val;

    switch(ctx->action_dlg.action_conf.joy.flags) {
      case CONTROLLER_ACTION_JOY_VAL_MIN:
      case CONTROLLER_ACTION_JOY_VAL_MAX:
      case CONTROLLER_ACTION_JOY_VAL_MDL:
      case CONTROLLER_ACTION_JOY_VAL_THROUGH:
      case CONTROLLER_ACTION_JOY_VAL_REVERSE:
          break;

      case CONTROLLER_ACTION_JOY_VAL_SPEC:
          EnableWindow(ctx->action_dlg.joy.ratio, TRUE);

          set_window_num_value(ctx->action_dlg.joy.ratio,
                               ctx->action_dlg.joy.ratio_spin,
                               ctx->action_dlg.action_conf.joy.val);
          break;

      case CONTROLLER_ACTION_JOY_VAL_RATIO:
          EnableWindow(ctx->action_dlg.joy.numer, TRUE);
          EnableWindow(ctx->action_dlg.joy.denom, TRUE);
          EnableWindow(ctx->action_dlg.joy.offst, TRUE);

          SetWindowText(ctx->action_dlg.joy.ratio, TEXT("val"));
          set_window_num_value(
              ctx->action_dlg.joy.numer,
              ctx->action_dlg.joy.numer_spin,
              ctx->action_dlg.action_conf.joy.ratio_numerator);
          set_window_num_value(
              ctx->action_dlg.joy.denom,
              ctx->action_dlg.joy.denom_spin,
              ctx->action_dlg.action_conf.joy.ratio_denominator);
          set_window_num_value(
              ctx->action_dlg.joy.offst,
              ctx->action_dlg.joy.offst_spin,
              ctx->action_dlg.action_conf.joy.offset);
          break;
    }
}


/* WM_INITDIALOG for apply */
static INT_PTR action_dlg_apply_init_dialog(HWND hwnd, UINT msg,
                                            WPARAM wparam, LPARAM lparam,
                                            prop_page_context_t *ctx)
{
    int i;
    struct uint_ptr_pair id_item_map[] = {
        {IDC_CHK_APPLY_MOD(1), &ctx->action_dlg.apply.mod[0]},
        {IDC_CHK_APPLY_MOD(2), &ctx->action_dlg.apply.mod[1]},
        {IDC_CHK_APPLY_MOD(3), &ctx->action_dlg.apply.mod[2]},
        {IDC_CHK_APPLY_MOD(4), &ctx->action_dlg.apply.mod[3]},
        {IDC_CHK_APPLY_MOD(5), &ctx->action_dlg.apply.mod[4]},
        {IDC_CHK_APPLY_MOD(6), &ctx->action_dlg.apply.mod[5]},
        {IDC_CHK_APPLY_MOD(7), &ctx->action_dlg.apply.mod[6]},
        {IDC_CHK_APPLY_MOD(8), &ctx->action_dlg.apply.mod[7]},

        {IDC_CHK_APPLY_MOD(9), &ctx->action_dlg.apply.mod[8]},
        {IDC_CHK_APPLY_MOD(10), &ctx->action_dlg.apply.mod[9]},
        {IDC_CHK_APPLY_MOD(11), &ctx->action_dlg.apply.mod[10]},
        {IDC_CHK_APPLY_MOD(12), &ctx->action_dlg.apply.mod[11]},
        {IDC_CHK_APPLY_MOD(13), &ctx->action_dlg.apply.mod[12]},
        {IDC_CHK_APPLY_MOD(14), &ctx->action_dlg.apply.mod[13]},
        {IDC_CHK_APPLY_MOD(15), &ctx->action_dlg.apply.mod[14]},
        {IDC_CHK_APPLY_MOD(16), &ctx->action_dlg.apply.mod[15]},

        {IDC_CMB_APPLY_ACTION, &ctx->action_dlg.apply.action},

        {0, NULL}
    };

    get_dialog_item(hwnd, id_item_map);

    /* set checkbox */
    {
        USHORT mod, mod_mask;

        if(ctx->action_dlg.action_conf.type == CONTROLLER_ACTION_APPLY) {
            mod = ctx->action_dlg.action_conf.apply.mod;
            mod_mask = ctx->action_dlg.action_conf.apply.mod_mask;
        } else {
            mod = 0x0000;
            mod_mask = 0x0000;
        }

        set_modifier_bits_to_checkbox(ctx->action_dlg.apply.mod,
                                      mod, mod_mask);
    }

    /* set combobox */
    {
        WPARAM idx;

        for(i = 0; i < ctx->data->conf.num_action_list; i++) {
            SendMessage(ctx->action_dlg.apply.action, CB_INSERTSTRING,
                        i, (LPARAM)ctx->data->conf.action_list[i].name);
        }

        if(ctx->action_dlg.action_conf.type == CONTROLLER_ACTION_APPLY) {
            idx = ctx->action_dlg.action_conf.apply.action_idx;
        } else {
            idx = -1;
        }
        SendMessage(ctx->action_dlg.apply.action, CB_SETCURSEL, idx, 0);
    }

    return TRUE;
}

/* WM_ACTION_DLG_QUERY_VALID for apply */
static INT_PTR action_dlg_apply_query_valid(HWND hwnd, UINT msg,
                                            WPARAM wparam, LPARAM lparam,
                                            prop_page_context_t *ctx)
{
    int idx;

    ctx->action_dlg.action_conf.type = CONTROLLER_ACTION_APPLY;

    idx = SendMessage(ctx->action_dlg.apply.action, CB_GETCURSEL, 0, 0);
    if(idx == CB_ERR) {
        display_message_box(IDS_MSG_ERR_SELECT_ACTION, IDS_MSG_ERROR,
                            hwnd, MB_OK | MB_ICONSTOP);
        SetFocus(ctx->action_dlg.apply.action);

        SetWindowLong(hwnd, DWL_MSGRESULT, FALSE);
        return TRUE;
    }

    get_modifier_bits_from_checkbox(
        ctx->action_dlg.apply.mod,
        &ctx->action_dlg.action_conf.apply.mod,
        &ctx->action_dlg.action_conf.apply.mod_mask);
    ctx->action_dlg.action_conf.apply.action_idx = idx;

    SetWindowLong(hwnd, DWL_MSGRESULT, TRUE);
    return TRUE;
}


/* page proc alist */
struct uint_ptr_pair action_dlg[] = {
    {WM_INITDIALOG, action_dlg_init_dialog},
    {WM_DESTROY, action_dlg_destroy},
    {WM_COMMAND, window_command_proc},

    {0, NULL}
};

static struct uint_ptr_pair action_dlg_mod[] = {
    {WM_INITDIALOG, action_dlg_mod_init_dialog},
    {WM_ACTION_DLG_QUERY_VALID, action_dlg_mod_query_valid},

    {0, NULL}
};
static struct uint_ptr_pair action_dlg_delay[] = {
    {WM_INITDIALOG, action_dlg_delay_init_dialog},
    {WM_ACTION_DLG_QUERY_VALID, action_dlg_delay_query_valid},

    {0, NULL}
};
static struct uint_ptr_pair action_dlg_joy[] = {
    {WM_INITDIALOG, action_dlg_joy_init_dialog},
    {WM_ACTION_DLG_QUERY_VALID, action_dlg_joy_query_valid},
    {WM_DRAWITEM, action_dlg_joy_drawitem},
    {WM_COMMAND, window_command_proc},

    {0, NULL}
};
static struct uint_ptr_pair action_dlg_apply[] = {
    {WM_INITDIALOG, action_dlg_apply_init_dialog},
    {WM_ACTION_DLG_QUERY_VALID, action_dlg_apply_query_valid},

    {0, NULL}
};


/* WM_COMMAND proc */
static struct uint_ptr_pair ok_proc_map[] = {
    {BN_CLICKED, action_dlg_ok_clicked},

    {0, NULL}
};
static struct uint_ptr_pair cancel_proc_map[] = {
    {BN_CLICKED, action_dlg_cancel_clicked},

    {0, NULL}
};
static struct uint_ptr_pair type_proc_map[] = {
    {CBN_SELCHANGE, action_dlg_type_change},

    {0, NULL}
};
static struct uint_ptr_pair command_proc_map[] ={
    {IDOK, &ok_proc_map},
    {IDCANCEL, &cancel_proc_map},
    {IDC_CMB_ACTION_TYPE, &type_proc_map},

    {0, NULL}
};


/* WM_COMMAND for joy */
static struct uint_ptr_pair joy_type_proc[] = {
    {CBN_SELCHANGE, action_dlg_joy_type_changed},

    {0, NULL}
};
static struct uint_ptr_pair joy_command_proc_map[] = {
    {IDC_CMB_ACTION_JOY_TYPE, joy_type_proc},

    {0, NULL}
};
