/*
 * setting.c  -- device setting
 *
 * $Id: setting.c,v 1.19 2004/03/13 17:40:41 hos Exp $
 *
 */

#include <windows.h>
#include "main.h"


/* tree specification */
static tree_item_spec_t custom_tree_stick_sub_items[] = {
    {TEXT("left side"), NULL, ITEM_TYPE_AXIS_LEFT,
     CONTROLLER_ACTION_JOY_CODE_X, 0},
    {TEXT("right side"), NULL, ITEM_TYPE_AXIS_RIGHT,
     CONTROLLER_ACTION_JOY_CODE_X, 0},
    {TEXT("top side"), NULL, ITEM_TYPE_AXIS_TOP,
     CONTROLLER_ACTION_JOY_CODE_Y, 0},
    {TEXT("bottom side"), NULL, ITEM_TYPE_AXIS_BTM,
     CONTROLLER_ACTION_JOY_CODE_Y, 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_hat_sub_items[] = {
    {TEXT("left side"), NULL, ITEM_TYPE_AXIS_LEFT,
     CONTROLLER_ACTION_JOY_CODE_RX, 0},
    {TEXT("right side"), NULL, ITEM_TYPE_AXIS_RIGHT,
     CONTROLLER_ACTION_JOY_CODE_RX, 0},
    {TEXT("top side"), NULL, ITEM_TYPE_AXIS_TOP,
     CONTROLLER_ACTION_JOY_CODE_RY, 0},
    {TEXT("bottom side"), NULL, ITEM_TYPE_AXIS_BTM,
     CONTROLLER_ACTION_JOY_CODE_RY, 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_btn_a_items[] = {
    {TEXT("motion"), NULL, ITEM_TYPE_BTN_ANALOG_MOTION,
     CONTROLLER_ACTION_JOY_CODE_BTN(3), 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_btn_b_items[] = {
    {TEXT("motion"), NULL, ITEM_TYPE_BTN_ANALOG_MOTION,
     CONTROLLER_ACTION_JOY_CODE_BTN(4), 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_dpad1_items[] = {
    {TEXT("left"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT1_LEFT, 0},
    {TEXT("right"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT1_RIGHT, 0},
    {TEXT("top"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT1_TOP, 0},
    {TEXT("bottom"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT1_BOTTOM, 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_stick_items[] = {
    {TEXT("stick"), custom_tree_stick_sub_items, ITEM_TYPE_AXIS_HV,
     CONTROLLER_ACTION_JOY_CODE_X, CONTROLLER_ACTION_JOY_CODE_Y},
    {TEXT("hat"), custom_tree_hat_sub_items, ITEM_TYPE_AXIS_HV,
     CONTROLLER_ACTION_JOY_CODE_RX, CONTROLLER_ACTION_JOY_CODE_RY},

    {TEXT("TRIGGER"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(0), 0},
    {TEXT("FIRE C"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(1), 0},
    {TEXT("LAUNCH"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(2), 0},
    {TEXT("BUTTON A"), custom_tree_btn_a_items, ITEM_TYPE_BTN_ANALOG,
     CONTROLLER_ACTION_JOY_CODE_BTN(3), 0},
    {TEXT("BUTTON B"), custom_tree_btn_b_items, ITEM_TYPE_BTN_ANALOG,
     CONTROLLER_ACTION_JOY_CODE_BTN(4), 0},
    {TEXT("HAT"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(5), 0},
    {TEXT("SW 1"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(6), 0},

    {TEXT("D-PAD 1"), custom_tree_dpad1_items, 0, 0, 0, TVIS_EXPANDED},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_throttle_sub_items[] = {
    {TEXT("top side"), NULL, ITEM_TYPE_AXIS_TOP,
     CONTROLLER_ACTION_JOY_CODE_THROTTLE, 0},
    {TEXT("bottom side"), NULL, ITEM_TYPE_AXIS_BTM,
     CONTROLLER_ACTION_JOY_CODE_THROTTLE, 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_rudder_sub_items[] = {
    {TEXT("left side"), NULL, ITEM_TYPE_AXIS_LEFT,
     CONTROLLER_ACTION_JOY_CODE_RZ, 0},
    {TEXT("right side"), NULL, ITEM_TYPE_AXIS_RIGHT,
     CONTROLLER_ACTION_JOY_CODE_RZ, 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_dpad2_items[] = {
    {TEXT("left"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT2_LEFT, 0},
    {TEXT("right"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT2_RIGHT, 0},
    {TEXT("top"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT2_TOP, 0},
    {TEXT("bottom"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_HAT2_BOTTOM, 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_dpad3_items[] = {
    {TEXT("left"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(12), 0},
    {TEXT("middle"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(13), 0},
    {TEXT("right"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(14), 0},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_throttle_items[] = {
    {TEXT("throttle"), custom_tree_throttle_sub_items, ITEM_TYPE_AXIS_V,
     CONTROLLER_ACTION_JOY_CODE_THROTTLE, 0},
    {TEXT("rudder"), custom_tree_rudder_sub_items, ITEM_TYPE_AXIS_H,
     CONTROLLER_ACTION_JOY_CODE_RZ, 0},

    {TEXT("BUTTON D"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(7), 0},
    {TEXT("BUTTON ST"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(8), 0},
    {TEXT("M1"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(9), 0},
    {TEXT("M2"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(10), 0},
    {TEXT("M3"), NULL, ITEM_TYPE_BTN_DIGITAL,
     CONTROLLER_ACTION_JOY_CODE_BTN(11), 0},

    {TEXT("D-PAD 2"), custom_tree_dpad2_items, 0, 0, 0, TVIS_EXPANDED},
    {TEXT("D-PAD 3"), custom_tree_dpad3_items, 0, 0, 0, TVIS_EXPANDED},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_controller_items[] = {
    {TEXT("stick block"), custom_tree_stick_items,
     0, 0, 0, TVIS_EXPANDED},
    {TEXT("throttle block"), custom_tree_throttle_items,
     0, 0, 0, TVIS_EXPANDED},

    {NULL, NULL, 0, 0, 0}
};

static tree_item_spec_t custom_tree_items[] = {
    /* controller setting */
    {TEXT("controller"), custom_tree_controller_items,
     ITEM_TYPE_CONTRLLER_ROOT, 0, 0, TVIS_EXPANDED},
    {TEXT("action list"), NULL, ITEM_TYPE_ACTION_ROOT, 0, 0, TVIS_EXPANDED},

    {NULL, NULL, 0, 0, 0}
};

tree_item_spec_t custom_tree_root = {
    /* root */
    TEXT("customize"), custom_tree_items, ITEM_TYPE_ROOT, 0, 0, TVIS_EXPANDED
};


int output_to_device(prop_common_data_t *comm_ctx,
                     const struct controller_output_report *rep)
{
    unsigned char buf[1 + sizeof(struct controller_output_report)];
    DWORD cnt;
    int ret;

    buf[0] = 0;                 /* collection 0 */
    memcpy(&buf[1], rep, sizeof(struct controller_output_report));

    ret = WriteFile(comm_ctx->hid_device, buf, sizeof(buf), &cnt, NULL);

    return ret;
}


static HKEY open_registry_subkey(LPWSTR subkey)
{
    HKEY base_key, key;
    int ret;

    /* open base registry path */
    ret = RegOpenKeyExW(HKEY_LOCAL_MACHINE, REGISTRY_REL_PATH, 0, KEY_READ,
                        &base_key);
    if(ret != ERROR_SUCCESS) {
        return NULL;
    }

    /* open subkey */
    ret = RegOpenKeyExW(base_key, subkey, 0, KEY_READ, &key);
    if(ret != ERROR_SUCCESS) {
        key = NULL;
    }

    RegCloseKey(base_key);

    return key;
}

static int alloc_copy_data(PVOID dst, PULONG dst_size,
                           PVOID src, ULONG src_size)
{
    void **dst_ptr;

    dst_ptr = (void **)dst;

    if(src_size == 0) {
        *dst_ptr = NULL;
        *dst_size = 0;
        return TRUE;
    }

    *dst_ptr = malloc(src_size);
    if(*dst_ptr == NULL) {
        *dst_size = 0;
        return FALSE;
    }

    memcpy(*dst_ptr, src, src_size);
    *dst_size = src_size;

    return TRUE;
}

static int read_registry_data(LPWSTR subkey, LPWSTR name,
                              PVOID val, PULONG size,
                              PVOID def_val, ULONG def_size, DWORD req_type)
{
    HKEY key;
    DWORD type, rsize;
    void *rval = NULL;
    int ret;

    /* open subkey */
    key = open_registry_subkey(subkey);
    if(key == NULL) {
        alloc_copy_data(val, size, def_val, def_size);
        ret = FALSE;
        goto func_end;
    }

    /* get size of data */
    rsize = 0;
    ret = RegQueryValueExW(key, name, NULL, &type, NULL, &rsize);
    if(ret != ERROR_SUCCESS || type != req_type) {
        alloc_copy_data(val, size, def_val, def_size);
        ret = FALSE;
        goto func_end;
    }

    rval = malloc(rsize);
    if(rval == NULL) {
        alloc_copy_data(val, size, def_val, def_size);
        ret = FALSE;
        goto func_end;
    }

    /* get data */
    ret = RegQueryValueExW(key, name, NULL, NULL, rval, &rsize);
    if(ret != ERROR_SUCCESS) {
        alloc_copy_data(val, size, def_val, def_size);
        ret = FALSE;
        goto func_end;
    }

    alloc_copy_data(val, size, rval, rsize);
    ret = TRUE;

  func_end:
    if(key != NULL) RegCloseKey(key);
    if(rval != NULL) free(rval);

    return ret;
}

static int read_registry_dword(LPWSTR subkey, LPWSTR name,
                               PDWORD val, DWORD def_val)
{
    DWORD *data;
    ULONG size;
    int ret;

    ret = read_registry_data(subkey, name,
                             &data, &size, &def_val, sizeof(DWORD), REG_DWORD);
    *val = *data;
    free(data);

    return ret;
}

static int read_registry_binary(LPWSTR subkey, LPWSTR name,
                                PVOID val, PULONG size,
                                PVOID def_val, ULONG def_size)
{
    return read_registry_data(subkey, name,
                              val, size, def_val, def_size, REG_BINARY);
}

static int read_registry_string(LPWSTR subkey, LPWSTR name,
                                LPWSTR *val, PULONG size,
                                LPWSTR def_val, ULONG def_size)
{
    return read_registry_data(subkey, name,
                              val, size, def_val, def_size, REG_SZ);
}


static HKEY create_registry_subkey(LPWSTR subkey)
{
    HKEY base_key, key;
    DWORD dispos;
    int ret;

    /* open base registry path */
    ret = RegCreateKeyExW(HKEY_LOCAL_MACHINE, REGISTRY_REL_PATH, 0, NULL, 0,
                          KEY_WRITE, NULL, &base_key, &dispos);
    if(ret != ERROR_SUCCESS) {
        return NULL;
    }

    /* open subkey */
    ret = RegCreateKeyExW(base_key, subkey, 0, NULL, 0,
                          KEY_WRITE, NULL, &key, &dispos);
    if(ret != ERROR_SUCCESS) {
        key = NULL;
    }

    RegCloseKey(base_key);

    return key;
}

static int write_registry_data(LPWSTR subkey, LPWSTR name,
                               LPVOID val, ULONG size, DWORD type)
{
    HKEY key;
    int ret;

    /* open subkey */
    key = create_registry_subkey(subkey);
    if(key == NULL) {
        return FALSE;
    }

    /* set value */
    ret = RegSetValueExW(key, name, 0, type, (BYTE *)val, size);
    if(ret != ERROR_SUCCESS) {
        ret = FALSE;
    } else {
        ret = TRUE;
    }

    RegCloseKey(key);

    return ret;
}

static int write_registry_dword(LPWSTR subkey, LPWSTR name, DWORD val)
{
    return write_registry_data(subkey, name, &val, sizeof(val), REG_DWORD);
}

static int write_registry_binary(LPWSTR subkey, LPWSTR name,
                                 LPVOID val, ULONG size)
{
    return write_registry_data(subkey, name, val, size, REG_BINARY);
}

static int write_registry_string(LPWSTR subkey, LPWSTR name,
                                 LPWSTR val, ULONG size)
{
    return write_registry_data(subkey, name, val, size, REG_SZ);
}


#define NUM_OF_DEF_ACTION \
  (1 + 6 + (4 + 4 + INPUT_REPORT_NUM_BUTTON) * 2)

static
struct controller_modifier_config default_motion_mod[CONTROLLER_EVENT_NUM];
static
struct controller_modifier_config default_press_mod[CONTROLLER_EVENT_NUM];
static
struct controller_modifier_config default_release_mod[CONTROLLER_EVENT_NUM];
static unsigned short default_action_list[NUM_OF_DEF_ACTION];
static struct controller_action_config default_action[NUM_OF_DEF_ACTION];
static LPWSTR default_action_name[] = {
    L"null action",
    L"axis X motion",
    L"axis Y motion",
    L"axis RX motion",
    L"axis RY motion",
    L"axis RZ motion",
    L"axis Throttle motion",
    L"POV 1 top press",
    L"POV 1 top release",
    L"POV 1 right press",
    L"POV 1 right release",
    L"POV 1 bottom press",
    L"POV 1 bottom release",
    L"POV 1 left press",
    L"POV 1 left release",
    L"POV 2 top press",
    L"POV 2 top release",
    L"POV 2 right press",
    L"POV 2 right release",
    L"POV 2 bottom press",
    L"POV 2 bottom release",
    L"POV 2 left press",
    L"POV 2 left release",
    L"button 1 press",
    L"button 1 release",
    L"button 2 press",
    L"button 2 release",
    L"button 3 press",
    L"button 3 release",
    L"button 4 press",
    L"button 4 release",
    L"button 5 press",
    L"button 5 release",
    L"button 6 press",
    L"button 6 release",
    L"button 7 press",
    L"button 7 release",
    L"button 8 press",
    L"button 8 release",
    L"button 9 press",
    L"button 9 release",
    L"button 10 press",
    L"button 10 release",
    L"button 11 press",
    L"button 11 release",
    L"button 12 press",
    L"button 12 release",
    L"button 13 press",
    L"button 13 release",
    L"button 14 press",
    L"button 14 release",
    L"button 15 press",
    L"button 15 release",
    L"button 16 press",
    L"button 16 release",
};
static LPWSTR default_profile_name = L"default profile";

static void generate_default_setting(void)
{
    int i, ai;

    memset(default_motion_mod, 0, sizeof(default_motion_mod));
    memset(default_press_mod, 0, sizeof(default_press_mod));
    memset(default_release_mod, 0, sizeof(default_release_mod));
    memset(default_action_list, 0, sizeof(default_action_list));
    memset(default_action, 0, sizeof(default_action));

    ai = 1;
    for(i = 0; i < CONTROLLER_EVENT_NUM; i++) {
        if(i <= CONTROLLER_ACTION_JOY_CODE_THROTTLE) {
            /* motion action for axis */
            default_action_list[ai] = ai;
            default_motion_mod[i].action_idx = ai;
            default_action[ai].type = CONTROLLER_ACTION_JOY;
            default_action[ai].joy.flags = CONTROLLER_ACTION_JOY_VAL_THROUGH;
            default_action[ai].joy.code = i;
            ai += 1;
        } else {
            /* press/release action for d-pad/digital button */
            default_action_list[ai] = ai;
            default_press_mod[i].action_idx = ai;
            default_action[ai].type = CONTROLLER_ACTION_JOY;
            default_action[ai].joy.flags = CONTROLLER_ACTION_JOY_VAL_MAX;
            default_action[ai].joy.code = i;
            ai += 1;

            default_action_list[ai] = ai;
            default_release_mod[i].action_idx = ai;
            default_action[ai].type = CONTROLLER_ACTION_JOY;
            default_action[ai].joy.flags = CONTROLLER_ACTION_JOY_VAL_MIN;
            default_action[ai].joy.code = i;
            ai += 1;
        }
    }
}


enum {
    CONF_LOAD,
    CONF_DFLT,
    CONF_SAVE
};

static int apply_device_setting(prop_common_data_t *comm_ctx, int load_flag)
{
    WCHAR subkey[256];
    DWORD map;
    struct {
        PWSTR name;
        PDWORD ptr;
        DWORD def_val;
    } name_ptr_list[] = {
        {L"button A max", &comm_ctx->conf.btn_a_sensitivity, 0x80},
        {L"button B max", &comm_ctx->conf.btn_b_sensitivity, 0x80},
        {L"button threshold", &comm_ctx->conf.btn_threshold, 0x80},
        {L"lower threshold", &comm_ctx->conf.lower_threshold, 0x10},
        {L"upper threshold", &comm_ctx->conf.upper_threshold, 0xef},
        {L"initial modifier", &comm_ctx->conf.initial_modifier, 0x0000},

        {NULL, NULL, 0}
    };
    int i;

    switch(load_flag) {
      case CONF_LOAD:
          read_registry_dword(L"map", L"current setting", &map, 1);
          break;

      case CONF_DFLT:
          map = 1;
          break;

      case CONF_SAVE:
          if(! read_registry_dword(L"map", L"current setting", &map, 1)) {
              /* save default map */
              write_registry_dword(L"map", L"current setting", map);
          }
          break;
    }

    wsprintfW(subkey, L"parameters\\%x", map);

    /* load/save DWORD setting from registry */
    for(i = 0; name_ptr_list[i].name != NULL; i++) {
        switch(load_flag) {
          case CONF_LOAD:
              /* load */
              read_registry_dword(subkey, name_ptr_list[i].name,
                                  name_ptr_list[i].ptr,
                                  name_ptr_list[i].def_val);
              break;

          case CONF_DFLT:
              *name_ptr_list[i].ptr = name_ptr_list[i].def_val;
              break;

          case CONF_SAVE:
              /* save */
              write_registry_dword(subkey, name_ptr_list[i].name,
                                   *name_ptr_list[i].ptr);
        }
    }

    /* load/save event config */
    {
        WCHAR name[256];
        ULONG size, def_size;
        unsigned short *act_list;
        int i, nact;

        struct {
            PWSTR name;
            struct controller_event_config *event;
            struct controller_modifier_config *def_mod;
        } *lptr, evt_name_list[] = {
            {L"motion ", comm_ctx->conf.motion_event, default_motion_mod},
            {L"press ", comm_ctx->conf.press_event, default_press_mod},
            {L"release ", comm_ctx->conf.release_event, default_release_mod},
            {L"lower in ", comm_ctx->conf.lower_in_event, NULL},
            {L"lower out ", comm_ctx->conf.lower_out_event, NULL},
            {L"upper in ", comm_ctx->conf.upper_in_event, NULL},
            {L"upper out ", comm_ctx->conf.upper_out_event, NULL},

            {NULL, NULL, NULL}
        };

        /* load/save modifier/action assoc list of each event */
        {
            struct controller_modifier_config *mod, *def_mod;
            int nmod;

            for(lptr = evt_name_list; lptr->name != NULL; lptr++) {
                for(i = 0; i < CONTROLLER_EVENT_NUM; i++) {
                    /* name of value */
                    wsprintfW(name, L"%s%x", lptr->name, i);

                    switch(load_flag) {
                      case CONF_LOAD:
                      case CONF_DFLT:
                          /* load modifier list */
                          size = 0;
                          def_mod = ((lptr->def_mod != NULL &&
                                      lptr->def_mod[i].action_idx != 0) ?
                                     &lptr->def_mod[i] : NULL);
                          def_size =
                              (def_mod != NULL ?
                               sizeof(struct controller_modifier_config) : 0);

                          if(load_flag == CONF_LOAD) {
                              read_registry_binary(subkey, name,
                                                   &mod, &size,
                                                   def_mod, def_size);
                          } else {
                              alloc_copy_data(&mod, &size, def_mod, def_size);
                          }

                          nmod = size /
                                 sizeof(struct controller_modifier_config);
                          lptr->event[i].num_mod = nmod;
                          lptr->event[i].mod = mod;
                          break;

                      case CONF_SAVE:
                          /* save modifier list */
                          mod = lptr->event[i].mod;
                          size = lptr->event[i].num_mod *
                                 sizeof(struct controller_modifier_config);

                          write_registry_binary(subkey, name,
                                                mod, size);
                          break;
                    }
                }
            }
        }

        /* load/save list of action list */
        {
            wsprintfW(name, L"action list");

            nact = 0;
            switch(load_flag) {
              case CONF_LOAD:
              case CONF_DFLT:
                  /* load */
                  if(load_flag == CONF_LOAD) {
                      read_registry_binary(subkey, name,
                                           &act_list, &size,
                                           default_action_list,
                                           sizeof(default_action_list));
                  } else {
                      alloc_copy_data(&act_list, &size,
                                      default_action_list,
                                      sizeof(default_action_list));
                  }

                  nact = size / sizeof(unsigned short);
                  comm_ctx->conf.num_action_list = nact;
                  break;

              case CONF_SAVE:
                  /* save */
                  nact = comm_ctx->conf.num_action_list;
                  size = nact * sizeof(unsigned short);
                  act_list = (unsigned short *)malloc(size);
                  if(act_list == NULL) {
                      return FALSE;
                  }
                  for(i = 0; i < nact; i++) {
                      act_list[i] = i;
                  }

                  write_registry_binary(subkey, name,
                                        act_list, size);
                  break;
            }
        }

        /* load/save action list, action name */
        {
            struct controller_action_config *act, *def_act;
            LPWSTR act_name, def_act_name;

            if(load_flag == CONF_LOAD || load_flag == CONF_DFLT) {
                /* setup list of action list */
                comm_ctx->conf.action_list =
                    (struct controller_action_config_list *)
                    malloc(sizeof(struct controller_action_config_list) *
                           nact);
                if(comm_ctx->conf.action_list == NULL) {
                    free(act_list);
                    return FALSE;
                }
            }

            for(i = 0; i < nact; i++) {
                /* action list */
                wsprintfW(name, L"action list %x", act_list[i]);

                switch(load_flag) {
                  case CONF_LOAD:
                  case CONF_DFLT:
                      /* load */
                      def_act =
                          ((i < NUM_OF_DEF_ACTION &&
                            default_action[default_action_list[i]].type !=
                            CONTROLLER_ACTION_NOP) ?
                           &default_action[default_action_list[i]] :
                           NULL);
                      def_size = (def_act != NULL ?
                                  sizeof(struct controller_action_config) : 0);

                      if(load_flag == CONF_LOAD) {
                          read_registry_binary(subkey, name,
                                               &act, &size,
                                               def_act, def_size);
                      } else {
                          alloc_copy_data(&act, &size, def_act, def_size);
                      }

                      comm_ctx->conf.action_list[i].num_action =
                          size / sizeof(struct controller_action_config);
                      comm_ctx->conf.action_list[i].action = act;
                      break;

                  case CONF_SAVE:
                      /* save */
                      act = comm_ctx->conf.action_list[i].action;
                      size = comm_ctx->conf.action_list[i].num_action *
                             sizeof(struct controller_action_config);

                      write_registry_binary(subkey, name,
                                            act, size);
                      break;
                }

                /* action name */
                wsprintfW(name, L"action name %x", act_list[i]);

                switch(load_flag) {
                  case CONF_LOAD:
                  case CONF_DFLT:
                      /* load */
                      def_act_name =
                          (i < NUM_OF_DEF_ACTION ?
                           default_action_name[default_action_list[i]] : NULL);
                      def_size =
                          (def_act_name != NULL ?
                           (lstrlenW(def_act_name) + 1) * sizeof(WCHAR) : 0);

                      if(load_flag == CONF_LOAD) {
                          read_registry_string(subkey, name,
                                               &act_name, &size,
                                               def_act_name, def_size);
                      } else {
                          alloc_copy_data(&act_name, &size,
                                          def_act_name, def_size);
                      }

                      comm_ctx->conf.action_list[i].name = act_name;
                      break;

                  case CONF_SAVE:
                      /* save */
                      act_name = comm_ctx->conf.action_list[i].name;
                      size = (lstrlenW(act_name) + 1) * sizeof(WCHAR);

                      write_registry_string(subkey, name,
                                            act_name, size);
                      break;
                }
            }
        }

        free(act_list);
    }

    /* load/save profile name */
    {
        ULONG size, def_size;
        LPWSTR name;

        name = L"profile name";

        switch(load_flag) {
          case CONF_LOAD:
              def_size = (lstrlenW(default_profile_name) + 1) * sizeof(WCHAR);
              read_registry_string(subkey, name,
                                   &comm_ctx->conf.profile_name, &size,
                                   default_profile_name, def_size);
              break;

          case CONF_DFLT:
              def_size = (lstrlenW(default_profile_name) + 1) * sizeof(WCHAR);
              alloc_copy_data(&comm_ctx->conf.profile_name, &size,
                              default_profile_name, def_size);
              break;

          case CONF_SAVE:
              size = (lstrlenW(comm_ctx->conf.profile_name) + 1) *
                     sizeof(WCHAR);
              write_registry_string(subkey, name,
                                    comm_ctx->conf.profile_name, size);
              break;
        }
    }

    return TRUE;
}

int load_device_setting(prop_common_data_t *comm_ctx)
{
    generate_default_setting();
    free_device_setting(comm_ctx);
    return ((apply_device_setting(comm_ctx, CONF_LOAD) &&
             verify_device_setting(comm_ctx)) ||
            load_default_setting(comm_ctx));
}

int load_default_setting(prop_common_data_t *comm_ctx)
{
    generate_default_setting();
    free_device_setting(comm_ctx);
    return apply_device_setting(comm_ctx, CONF_DFLT);
}

int save_device_setting(prop_common_data_t *comm_ctx)
{
    return apply_device_setting(comm_ctx, CONF_SAVE);
}

void free_device_setting(prop_common_data_t *comm_ctx)
{
    int i;

#define FREE_IF_NON_NULL(_p_) if((_p_) != NULL) { free(_p_); (_p_) = NULL; }
#define FREE_MOD_IF_NON_NULL(_p_)               \
  FREE_IF_NON_NULL((_p_).mod);                  \
  (_p_).num_mod = 0;

    for(i = 0; i < CONTROLLER_EVENT_NUM; i++) {
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.motion_event[i]);
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.press_event[i]);
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.release_event[i]);
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.lower_in_event[i]);
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.lower_out_event[i]);
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.upper_in_event[i]);
        FREE_MOD_IF_NON_NULL(comm_ctx->conf.upper_out_event[i]);
    }

    for(i = 0; i < comm_ctx->conf.num_action_list; i++) {
        FREE_IF_NON_NULL(comm_ctx->conf.action_list[i].action);
        FREE_IF_NON_NULL(comm_ctx->conf.action_list[i].name);
        comm_ctx->conf.action_list[i].num_action = 0;
    }
    FREE_IF_NON_NULL(comm_ctx->conf.action_list);
    comm_ctx->conf.num_action_list = 0;

#undef FREE_MOD_IF_NON_NULL
#undef FREE_IF_NON_NULL

    memset(&comm_ctx->conf, 0, sizeof(comm_ctx->conf));
}

int verify_device_setting(prop_common_data_t *comm_ctx)
{
    int i, j, k;

    /* range of DWORD values */
    {
        struct {
            DWORD *val;
            DWORD min;
            DWORD max;
        } val_range[] = {
            {&comm_ctx->conf.btn_a_sensitivity, 0x00, 0xff},
            {&comm_ctx->conf.btn_b_sensitivity, 0x00, 0xff},
            {&comm_ctx->conf.btn_threshold, 0x00, 0xff},
            {&comm_ctx->conf.lower_threshold, 0x00, 0xff},
            {&comm_ctx->conf.upper_threshold, 0x00, 0xff},
            {&comm_ctx->conf.initial_modifier, 0x00, 0xff},

            {NULL, 0, 0}
        };

        for(i = 0; val_range[i].val != NULL; i++) {
            if(*val_range[i].val < val_range[i].min ||
               *val_range[i].val > val_range[i].max) {
                return FALSE;
            }
        }
    }

    /* events */
    {
        struct controller_event_config *event[] = {
            comm_ctx->conf.motion_event,
            comm_ctx->conf.press_event,
            comm_ctx->conf.release_event,
            comm_ctx->conf.lower_in_event,
            comm_ctx->conf.lower_out_event,
            comm_ctx->conf.upper_in_event,
            comm_ctx->conf.upper_out_event,

            NULL
        };

        for(i = 0; event[i] != NULL; i++) {
            for(j = 0; j < CONTROLLER_EVENT_NUM; j++) {
                for(k = 0; k < event[i][j].num_mod; k++) {
                    if(event[i][j].mod[k].action_idx >=
                       comm_ctx->conf.num_action_list) {
                        return FALSE;
                    }
                }
            }
        }
    }

    /* action list */
    for(i = 0; i < comm_ctx->conf.num_action_list; i++) {
        if(lstrlenW(comm_ctx->conf.action_list[i].name) <= 0) {
            return FALSE;
        }

        for(j = 0; j < comm_ctx->conf.action_list[i].num_action; j++) {
            if(comm_ctx->conf.action_list[i].action[j].type ==
               CONTROLLER_ACTION_APPLY &&
               comm_ctx->conf.action_list[i].action[j].apply.action_idx >=
               comm_ctx->conf.num_action_list) {
                return FALSE;
            }
        }
    }

    return TRUE;
}
