/*
 * hid.c  -- HID (Human Interface Device)
 *
 * $Id: hid.c,v 1.24 2004/03/13 16:11:46 hos Exp $
 *
 */


#include "main.h"


#define REPORT_DESCRIPTOR_SIZE (sizeof(hid_report_descriptor))

static BYTE hid_report_descriptor[] = {
    /* joystick */
    0x05, 0x01,                 /* Usage Page(Generic Desktop) */
    0x09, 0x04,                 /* Usage(Joystick) */
    0xa1, 0x01,                 /* Collection(Application) */

    /* x, y */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x02,                 /*     Report Count(2) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x30,                 /*     Usage(X) */
    0x09, 0x31,                 /*     Usage(Y) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* rx, ry */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x02,                 /*     Report Count(2) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x33,                 /*     Usage(Rx) */
    0x09, 0x34,                 /*     Usage(Ry) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* rz */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x01,                 /*     Report Count(1) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x35,                 /*     Usage(Rz) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* throttle */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x05, 0x02,                 /*   Usage Page(Simulation) */
    0x09, 0xbb,                 /*   Usage(Throttle) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* hat switch 1 */
    0x75, 0x04,                 /*   Report Size(4) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x07,                 /*   Logical Maximum(0x07) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0x3b, 0x01,           /*   Physical Maximum(0x013b) */
    0x65, 0x14,                 /*   Unit(Degree) */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x39,                 /*   Usage(Hat Switch) */
    0x81, 0x42,                 /*   Input(Variable, Absolute, Null State) */

    /* hat switch 2 */
    0x75, 0x04,                 /*   Report Size(4) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x07,                 /*   Logical Maximum(0x07) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0x3b, 0x01,           /*   Physical Maximum(0x013b) */
    0x65, 0x14,                 /*   Unit(Degree) */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x39,                 /*   Usage(Hat Switch) */
    0x81, 0x42,                 /*   Input(Variable, Absolute, Null State) */

    /* button x 16 */
    0x75, 0x01,                 /*   Report Size(1) */
    0x95, 0x10,                 /*   Report Count(16) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x01,                 /*   Logical Maximum(0x01) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x01,                 /*   Physical Maximum(0x01) */
    0x05, 0x09,                 /*   Usage Page(Button) */
    0x19, 0x01,                 /*   Usage Minimum(Button 1) */
    0x29, 0x10,                 /*   Usage Maximum(Button 16) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* modifier state (16bits) */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x02,                 /*   Report Count(2) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x06, 0x00, 0xff,           /*   Usage Page(Vendor defined) */
    0x09, 0x10,                 /*   Usage(Vendor defined 16) */
    0x09, 0x11,                 /*   Usage(Vendor defined 17) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* Output Report */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x06, 0x00, 0xff,           /*   Usage Page(Vendor defined) */
    0x09, 0x01,                 /*   Usage(Vendor defined 1) */
    0x91, 0x02,                 /*   Output(Variable, Absolute) */

    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x06, 0x00, 0xff,           /*   Usage Page(Vendor defined) */
    0x09, 0x02,                 /*   Usage(Vendor defined 2) */
    0x91, 0x02,                 /*   Output(Variable, Absolute) */

    0xc0,                       /* End Collection */
};


static VOID hid_update_from_raw_input(IN PDEVICE_OBJECT dev_obj);
static VOID hid_generate_raw_output(struct controller_output_report *out_state,
                                    struct controller_raw_output *raw_out);
static VOID DDKAPI hid_worker(IN PVOID ctxt);


NTSTATUS hid_start(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;

    DBG_ENTER(("%p", dev_obj));

    dev_ext = GET_DEV_EXT(dev_obj);

    /* initialize hid part */
    dev_ext->report_size = REPORT_DESCRIPTOR_SIZE;
    dev_ext->firstp = TRUE;
    RtlZeroMemory(&dev_ext->raw_input, sizeof(dev_ext->raw_input));
    RtlZeroMemory(&dev_ext->pp_input, sizeof(dev_ext->pp_input));
    RtlZeroMemory(&dev_ext->pr_input, sizeof(dev_ext->pr_input));
    RtlZeroMemory(&dev_ext->input_state, sizeof(dev_ext->input_state));

    /* start hid worker */
    {
        HANDLE thread;
        NTSTATUS nt_ret;

        /* create thread and start */
        nt_ret = PsCreateSystemThread(&thread,
                                      THREAD_ALL_ACCESS, NULL, NULL, NULL,
                                      hid_worker, dev_obj);
        if(! NT_SUCCESS(nt_ret)) {
            DBG_LEAVE(("%x: PsCreateSystemThread failed", nt_ret));
            return nt_ret;
        }

        ObReferenceObjectByHandle(thread, THREAD_ALL_ACCESS, NULL, KernelMode,
                                  &dev_ext->hid_worker_thread_obj, NULL);
    }

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}

NTSTATUS hid_remove(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;

    DBG_ENTER(("%p", dev_obj));

    dev_ext = GET_DEV_EXT(dev_obj);

    /* wait for hid worker stop */
    KeWaitForSingleObject(dev_ext->hid_worker_thread_obj,
                          Executive, KernelMode, FALSE, NULL);    

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS hid_get_rep_desc(IN PDEVICE_OBJECT dev_obj,
                          OUT BYTE *buf, OUT ULONG *size)
{
    DBG_ENTER(("%p, %p, %p", dev_obj, buf, size));

    if(*size < REPORT_DESCRIPTOR_SIZE) {
        DBG_LEAVE(("STATUS_BUFFER_TOO_SMALL"));
        return STATUS_BUFFER_TOO_SMALL;
    }

    RtlCopyMemory(buf, hid_report_descriptor, REPORT_DESCRIPTOR_SIZE);
    *size = REPORT_DESCRIPTOR_SIZE;

    DBG_LEAVE(("STATUS_SUCCESS: %u", *size));
    return STATUS_SUCCESS;
}


NTSTATUS hid_read_rep(IN PDEVICE_OBJECT dev_obj,
                      OUT BYTE *buf, OUT ULONG *size)
{
    device_extension_t *dev_ext;
    struct controller_input_report in_state;

    DBG_ENTER(("%p, %p, %u", dev_obj, buf, *size));

    dev_ext = GET_DEV_EXT(dev_obj);

    if(*size < sizeof(in_state)) {
        DBG_LEAVE(("STATUS_BUFFER_TOO_SMALL"));
        return STATUS_BUFFER_TOO_SMALL;
    }

    if(dev_ext->firstp) {
        /* first read report request: indicate to reload setting */
        dev_ext->reload_settingp = TRUE;
        dev_ext->firstp = FALSE;
    }

    in_state = dev_ext->input_state;

    RtlCopyMemory(buf, &in_state, sizeof(in_state));
    *size = sizeof(in_state);

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS hid_write_rep(IN PDEVICE_OBJECT dev_obj,
                       IN BYTE *buf, IN ULONG *size, IN UCHAR id)
{
    device_extension_t *dev_ext;
    struct controller_output_report out_state;
    NTSTATUS nt_ret;

    DBG_ENTER(("%p, %p, %u, %u", dev_obj, buf, *size, id));

    dev_ext = GET_DEV_EXT(dev_obj);
    nt_ret = STATUS_SUCCESS;

    if(*size > sizeof(out_state)) {
        *size = sizeof(out_state);
    }

    RtlZeroMemory(&out_state, sizeof(out_state));
    RtlCopyMemory(&out_state, buf, *size);

    switch(out_state.command) {
      case WRITE_REPORT_CMD_VIBLATION:
      {
          struct controller_raw_output raw_out;
          ULONG raw_size;

          hid_generate_raw_output(&out_state, &raw_out);

          DBG_OUT(("write report: vr0c: %02x (%02x)",
                   raw_out.vr_0c.viblation, out_state.value));
          raw_size = sizeof(raw_out.vr_0c);
          nt_ret = usb_put_vendor_request(dev_obj, &raw_out.vr_0c, &raw_size,
                                          0x0c, 0x01);
          break;
      }

      case WRITE_REPORT_CMD_RELOAD_SETTING:
          dev_ext->reload_settingp = TRUE;
          break;

      default:
          DBG_OUT(("unknown write command: 0x%x", out_state.command));
          break;
    }

    DBG_LEAVE(("%x", nt_ret));
    return nt_ret;
}


static VOID hid_update_from_raw_input(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;
    struct controller_raw_input *raw_in;
    struct controller_input_report in_state;

    DBG_ENTER2(("%p", dev_obj));

    dev_ext = GET_DEV_EXT(dev_obj);

    raw_in = &dev_ext->raw_input;
    in_state = dev_ext->input_state;

    /* process input data */
    process_raw_input(dev_obj, raw_in, &in_state);

    /* save current state */
    dev_ext->input_state = in_state;

    DBG_LEAVE2(("void"));
}

static VOID hid_generate_raw_output(struct controller_output_report *out_state,
                                    struct controller_raw_output *raw_out)
{
    DBG_ENTER(("%p, %p", out_state, raw_out));

    RtlZeroMemory(raw_out, sizeof(*raw_out));

    raw_out->vr_0c.viblation = out_state->value;

    DBG_LEAVE(("void"));
}


static VOID DDKAPI hid_worker(IN PVOID ctxt)
{
    PDEVICE_OBJECT dev_obj;
    device_extension_t *dev_ext;
    LARGE_INTEGER time;

    DBG_ENTER(("%p", dev_obj, ctxt));

    dev_obj = (PDEVICE_OBJECT)ctxt;
    dev_ext = GET_DEV_EXT(dev_obj);

    KeSetPriorityThread(KeGetCurrentThread(), LOW_REALTIME_PRIORITY);

    while(! dev_ext->removedp) {
        if(dev_ext->reload_settingp) {
            load_settings(dev_obj);
            dev_ext->reload_settingp = FALSE;
        }

        hid_update_from_raw_input(dev_obj);

        time.QuadPart = 10 * 1000 * -10;
        KeDelayExecutionThread(KernelMode, FALSE, &time);
    }

    DBG_OUT(("already removed"));

    /* free memory before exit */
    free_settings(dev_obj);

    DBG_LEAVE2(("STATUS_SUCCESS"));
    PsTerminateSystemThread(STATUS_SUCCESS);
}
