/*
 * report_data.h  -- define report data format 
 *
 * $Id: report_data.h,v 1.13 2004/03/13 16:11:47 hos Exp $
 *
 */


/* read report */

#define INPUT_REPORT_NUM_BUTTON 16

#pragma pack(push, 1)
struct controller_input_report {
    union {
        struct {
            UCHAR x;            /* X:  0x00 - 0xff */
            UCHAR y;            /* Y:  0x00 - 0xff */
            UCHAR rx;           /* RX: 0x00 - 0xff */
            UCHAR ry;           /* RY: 0x00 - 0xff */
            UCHAR rz;           /* RZ: 0x00 - 0xff */
            UCHAR throttle;     /* TH: 0x00 - 0xff */
        };

        struct {
            UCHAR axis[6];      /* x, y, rx, ry, yz, throttle */
        };
    };

    UCHAR hat1 : 4;             /* H1: 0x00 - 0x07 */
    UCHAR hat2 : 4;             /* H2: 0x00 - 0x07 */

    UCHAR button[INPUT_REPORT_NUM_BUTTON / 8]; /* BN: 0x00/0x01 */

    USHORT modifier;            /* modifier state */
};
#pragma pack(pop)


/* write report */

#define WRITE_REPORT_CMD_VIBLATION 0x01
#define WRITE_REPORT_CMD_RELOAD_SETTING 0x02

struct controller_output_report {
    UCHAR command;              /* command: 0x00 - 0xff */
    UCHAR value;                /* value: 0x00 - 0xff */
};


/* setting stored in registroy */
#define REGISTRY_REL_PATH L"Software\\noname\\hfsd"
#define REGISTRY_BASE_PATH L"\\Registry\\Machine\\" REGISTRY_REL_PATH


/* configuration of value processor */

#define CONTROLLER_EVENT_NUM 30
#define CONTROLLER_MODEFIER(n) ((USHORT)(1 << ((n) - 1)))

/* sizeof(struct controller_modifier_config) == 4 */
struct controller_modifier_config {
    USHORT mod;                 /* modifier */
    USHORT mod_mask;            /* modifier mask */

    USHORT action_idx;          /* index of action list */

    char reserved[2];
};


/* controller_action_config.type */
#define CONTROLLER_ACTION_NOP   0x00
#define CONTROLLER_ACTION_MOD   0x01
#define CONTROLLER_ACTION_DELAY 0x02
#define CONTROLLER_ACTION_JOY   0x03
#define CONTROLLER_ACTION_APPLY 0x04
#define CONTROLLER_ACTION_MOUSE 0x10
#define CONTROLLER_ACTION_KBD   0x11

/* controller_action_config.joy.flags */
#define CONTROLLER_ACTION_JOY_VAL_MIN     0x0001
#define CONTROLLER_ACTION_JOY_VAL_MAX     0x0002
#define CONTROLLER_ACTION_JOY_VAL_MDL     0x0003
#define CONTROLLER_ACTION_JOY_VAL_THROUGH 0x0004
#define CONTROLLER_ACTION_JOY_VAL_REVERSE 0x0005
#define CONTROLLER_ACTION_JOY_VAL_SPEC    0x0006
#define CONTROLLER_ACTION_JOY_VAL_RATIO   0x0007

/* controller_action_config.joy.code */
#define CONTROLLER_ACTION_JOY_CODE_X           0x00
#define CONTROLLER_ACTION_JOY_CODE_Y           0x01
#define CONTROLLER_ACTION_JOY_CODE_RX          0x02
#define CONTROLLER_ACTION_JOY_CODE_RY          0x03
#define CONTROLLER_ACTION_JOY_CODE_RZ          0x04
#define CONTROLLER_ACTION_JOY_CODE_THROTTLE    0x05

#define CONTROLLER_ACTION_JOY_CODE_HAT1_TOP    0x06
#define CONTROLLER_ACTION_JOY_CODE_HAT1_RIGHT  0x07
#define CONTROLLER_ACTION_JOY_CODE_HAT1_BOTTOM 0x08
#define CONTROLLER_ACTION_JOY_CODE_HAT1_LEFT   0x09

#define CONTROLLER_ACTION_JOY_CODE_HAT2_TOP    0x0a
#define CONTROLLER_ACTION_JOY_CODE_HAT2_RIGHT  0x0b
#define CONTROLLER_ACTION_JOY_CODE_HAT2_BOTTOM 0x0c
#define CONTROLLER_ACTION_JOY_CODE_HAT2_LEFT   0x0d

#define CONTROLLER_ACTION_JOY_CODE_BTN(n)      (0x0e + (n))

#define CONTROLLER_ACTION_JOY_CODE_MAX \
  CONTROLLER_ACTION_JOY_CODE_BTN(INPUT_REPORT_NUM_BUTTON - 1)

/* sizeof(struct controller_action_config) == 32 */
struct controller_action_config {
    UCHAR type;                 /* type of action */

    char reserved[3];

    union {
        /* modifier */
        struct {
            USHORT mod;         /* modifier */
            USHORT mod_mask;    /* modifier mask */
        } mod;

        /* delay */
        struct {
            ULONG msec;         /* millisecond */
        } delay;

        /* joystick */
        struct {
            USHORT flags;       /* flags */
            USHORT code;        /* code */
            UCHAR val;          /* optional value */
            char reserved[1];
            SHORT ratio_numerator; /* numerator of ratio */
            SHORT ratio_denominator; /* denominator of ratio */
            SHORT offset;       /* offset */
        } joy;

        /* apply other action */
        struct {
            USHORT mod;         /* modifier: apply if match */
            USHORT mod_mask;    /* modifier mask */
            USHORT action_idx;  /* action index */
        } apply;

        /* keyboard */
        struct {
            USHORT flags;       /* KEYBOARD_INPUT_DATA.Flags */
            USHORT code;        /* KEYBOARD_INPUT_DATA.MakeCode */
            ULONG extra_info;   /* KEYBOARD_INPUT_DATA.ExtraInformation */
        } kbd;

        /* mouse */
        struct {
            USHORT flags;        /* MOUSE_INPUT_DATA.Flags */
            USHORT button_flags; /* MOUSE_INPUT_DATA.ButtonFlags */
            USHORT button_data;  /* MOUSE_INPUT_DATA.ButtonData */
            char reserved[2];
            LONG x;              /* MOUSE_INPUT_DATA.LastX */
            LONG y;              /* MOUSE_INPUT_DATA.LastY */
            ULONG extra_info;    /* MOUSE_INPUT_DATA.ExtraInformation */
        } mouse;

        char place_holder[28];
    };
};
